/*
 * main.c for XGalaga
 * Copyright (c) 1995 Joe Rumsey
 */
#include "copyright.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <math.h> 
#include "libsprite/Wlib.h"
#include "data.h"
#include "images.h"
#include "defs.h"
#include "struct.h"
#include "paths.h"
#include "proto.h"
#include "sound.h"

struct torp torps[MAXTORPS];
struct star stars[NUMSTARS];
struct torp *first_etorp=0;
int cheat_6_shot= 0;
int cheat_7_shot= 0;

player_t player_0;

static double convoyx = 0, convoymove = 1;
static int livecount = 0;
static int starspeed = 1;
static int attacking = 0, maxattacking, entering=0;
int maxetorps = 5, numetorps=0;
static int killed_aliens=0;

/* #define convoy_x_pos(i) (convoyx+(20 * (i - 10 * (i/10))))*/
#define convoy_x_pos(i) (convoyx+(20 * (i%10)))
#define convoy_y_pos(i) (20 + (20*(i/10)))

static double moves[16][2] = {
  { 0,-4},
  { 1,-4},
  { 3,-3},
  { 4,-1},
  { 4, 0},
  { 4, 1},
  { 3, 3},
  { 1, 4},
  { 0, 4},
  {-1, 4},
  {-3, 3},
  {-4, 1},
  {-4, 0},
  {-4,-1},
  {-3,-3},
  {-1,-4}
};

void
  xgal_exit(int v)
  {
    /*    W_AutoRepeatOn();*/
#ifdef SOUND
    kill_sound();
#endif
    /*    sleep(1);*/ /* Without this, the auto-repeat request fails on my machine... */
    exit(v);
  }


static void print_usage()
  {
    printf("\
  XGalaga v%s\n\
  Copyright (c) 1995 Joe Rumsey\n\
  <ogre@atomic.com>\n\
  Command line options:\n\
  -scores              Prints out the high score files and exits\n\
  -display <display>   Set your display\n\
  -mouse               Use mouse control (same as 'm' at the title screen)\n\
  -keyboard            Use keyboard control (same as 'k')\n%s\
  -b                   turn buffered mode off, use this if it runs\n\
  too slowly.  Will cause flicker, maybe lots,\n\
  maybe only a little.\n\
  \n\
  This game has the following copyrights:\n\
  Copyright (c) 1995   Joe Rumsey\n\
  Permission to use, copy, modify, and distribute this\n\
  software and its documentation for any purpose and without\n\
  fee is hereby granted, provided that the above copyright\n\
  notice appear in all copies and that both that copyright\n\
  notice and this permission notice appear in supporting\n\
  documentation.  No representations are made about the\n\
  suitability of this software for any purpose.  It is\n\
  provided \"as is\" without express or implied warranty.\n\
  (From copyright.h)\n\
  Copyright 1989		Kevin P. Smith\n\
  Scott Silvey\n\
  Permission to use, copy, modify, and distribute this\n\
  software and its documentation for any purpose and without\n\
  fee is hereby granted, provided that the above copyright\n\
  notice appear in all copies.\n\
  (From copyright2.h)\n\
  __LLeweLLyn__Reese__ added some hacks to create more aliens\n\
  as you rise in level.\n\
  So if this game is too hard for you, well, you can blame me.\n\
  Basic instructions:\n\
  It's Galaga, you know how to play Galaga, stop bothering me.\n\
  (Ship follows the mouse, button fires.  Auto-fire by holding it\n\
  down, so no-one accuses me of breaking their mouse!)\n\
  \n\
  Keyboard commands:\n\
  \n\
  p - pauses\n\
  q - end this game\n\
  b - Toggle buffering (flicker vs. speed.)\n\
  o - Toggle OR mode.  Don't use this, it sucks.\n", VERSION,
#ifdef SOUND
      "  -nosound             Turn sound OFF\n"
#else
      ""
#endif
  );
  }

/*------------------stars-----------------*/
static void init_stars()
  {
    int i;

    for(i=0;i<NUMSTARS;i++) {
      stars[i].x = random()%WINWIDTH;
      stars[i].y = random()%WINHEIGHT;
      stars[i].speed = (random()%3)+1;
      switch(random()%5) {
        case 0:
          stars[i].color = W_White;
          break;
        case 1:
          stars[i].color = W_Green;
          break;
        case 2:
          stars[i].color = W_Cyan;
          break;
        case 3:
          stars[i].color = W_Red;
          break;
        default:
          stars[i].color = W_Yellow;
          break;
      }
    }
  }

static int drewlevel = 0;

static void undo_stars()
  {
    int i;
    
    for(i=0;i<NUMSTARS;i++) {
      W_DrawPoint(baseWin, (int)stars[i].x, (int)stars[i].y, W_Black);
    }
    if(drewlevel) {
      W_ClearArea
      (
        baseWin,
        WINWIDTH/2 - (W_StringWidth("LEVEL 000", W_BigFont)/2), WINHEIGHT/2-W_BigTextheight/2,
        10*W_BigTextwidth, W_BigTextheight
    );
      drewlevel = 0;
    }
  }

static void do_stars()
  {
    int i;
    
    for(i=0;i<NUMSTARS;i++) {
      if(!paused)
      stars[i].y+=stars[i].speed*((starspeed < 20) ? ABS(starspeed) : 20);
      if(stars[i].y >= WINHEIGHT) {
        stars[i].y-=WINHEIGHT+starspeed;
        stars[i].x = random() % WINWIDTH;
        switch(random()%5) {
          case 0:
            stars[i].color = W_White;
            break;
          case 1:
            stars[i].color = W_Green;
            break;
          case 2:
            stars[i].color = W_Cyan;
            break;
          case 3:
            stars[i].color = W_Red;
            break;
          default:
            stars[i].color = W_Yellow;
            break;
        }
      }
      W_DrawPoint(baseWin, stars[i].x, stars[i].y, stars[i].color);
    }
    if(starspeed != 1) {
      char buf[20];
    
      drewlevel = 1;
      sprintf(buf, "LEVEL %d", level+1);
      W_MaskText(baseWin, WINWIDTH/2 - (W_StringWidth(buf, W_BigFont)/2), WINHEIGHT/2-W_BigTextheight/2,
        W_Red, buf, strlen(buf), W_BigFont);
    }
  }

/*-------------------aliens---------------*/

static void delete_etorps()
  {
    struct torp *tmp;

    while(first_etorp) {
      tmp = first_etorp->next;
      free(first_etorp);
      first_etorp = tmp;
    }
  }

static void init_aliens(int level)
  {
    int i;

    convoyx=0;
    convoymove = 1;

    maxattacking = 1 + (level*2);
    if(maxattacking > 30)
    maxattacking = 30;
    attacking = 0;
    maxetorps = 10 + (level*5);
    numetorps = 0;

    delete_etorps();
    for(i=0;i<MAXALIENS;i++) {
      new_alien(level, i, &aliens[i]);
    }

#if defined(MORE_ALIENS)
    /*
      Leave 60+10*(level/10) aliens alive.
    */
    for(i=(ENTRY_DEFINED_ALIENS+MORE_ALIENS_PER_INC*(level/LEVELS_PER_INC));i<MAXALIENS;++i)
    {
      aliens[i].alive=0;
      aliens[i].state=DEAD;
    }
#endif /* defined(MORE_ALIENS)    */

#if !defined(MORE_FLAGSHIPS)    
    /*
      Knock all but 2 of the flagships.
    */
    aliens[0].state=aliens[0].alive = 0;
    aliens[1].state=aliens[1].alive = 0;
    aliens[2].state=aliens[2].alive = 0;
    aliens[4].state=aliens[4].alive = 0;
    aliens[5].state=aliens[5].alive = 0;
    aliens[7].state=aliens[7].alive = 0;
    aliens[8].state=aliens[8].alive = 0;
    aliens[9].state=aliens[9].alive = 0;

    /*
      Knock out the first 2 and last 2 guards.
    */
    aliens[10].state=aliens[10].alive = 0;
    aliens[11].state=aliens[11].alive = 0;
    aliens[18].state=aliens[18].alive = 0;
    aliens[19].state=aliens[19].alive = 0;

#else /*!defined(MORE_FLAGSHIPS)    */
    /*
      Knock most of the flagships, depending on level.
      Warning! this kludge depends on the fact that 0==DEAD, and 1==WAITING_TO_ENTER.
    */
    aliens[0].alive*=aliens[0].state=(level>=32);/* 10 flagships at level 32*/
    aliens[1].alive*=aliens[1].state=(level>=16);/* 8 flagships at level 16*/
    aliens[2].alive*=aliens[2].state=(level>=8); /* 6 flagships at level 8*/
    aliens[3].alive*=aliens[3].state=(level>=0); 
    aliens[4].alive*=aliens[4].state=(level>=4); /* 4 flagships at level 4*/
    aliens[5].alive*=aliens[5].state=(level>=4);
    aliens[6].alive*=aliens[6].state=(level>=0);
    aliens[7].alive*=aliens[7].state=(level>=8);
    aliens[8].alive*=aliens[8].state=(level>=16);
    aliens[9].alive*=aliens[9].state=(level>=32);

    /*
      Knock some guards, depending on level.
    */
    aliens[10].alive*=aliens[10].state=(level>=8);
    aliens[11].alive*=aliens[11].state=(level>=4);
    aliens[18].alive*=aliens[18].state=(level>=4);
    aliens[19].alive*=aliens[19].state=(level>=8);

    livecount=0;
    for(i=0;i<MAXALIENS;++i)
    {
      if(aliens[i].alive)
      {
        ++livecount;
      }
    }

#endif /*!defined(MORE_FLAGSHIPS)    */
    
    for(i=0;i<MAXTORPS;i++)
    torps[i].alive = 0;

    killed_aliens=0;
  }

    
static void undo_aliens()
  {
    int i;

    for(i=0;i<MAXALIENS;i++) {
      if(aliens[i].alive)
      W_ClearArea
      (
        baseWin, 
        (int)aliens[i].x-(aliens[i].shape->width/2), 
        (int)aliens[i].y-(aliens[i].shape->height/2),
        aliens[i].shape->width, aliens[i].shape->height
    );
      if(aliens[i].dying)
      {
        aliens[i].alive = 0;
        aliens[i].dying=0;
        aliens[i].state=DEAD;
      }
    }
  }

static void
  alien_escorting(alien_t* escort_ap)
  {
    int fs = escort_ap->escorting;
#if defined(DEBUG_ALIEN_ESCORTING)
    fprintf
    (
      OUT_FILE,
      "Alien %i escorting flagship %i\n",escort_ap->id,aliens[fs].id
  );
#endif defined(DEBUG_ALIEN_ESCORTING)  
    if(!aliens[fs].alive)
    {
      escort_ap->escorting = -1;
      escort_ap->state=ATTACKING;
    }
    else
    if(aliens[fs].state==ATTACKING)
    {
#if defined(DEBUG_ALIEN_ESCORTING)
      fprintf
      (
        OUT_FILE,
        "Alien %i escorting flagship %i setting dir to %i\n",escort_ap->id,aliens[fs].id,aliens[fs].dir
    );
#endif defined(DEBUG_ALIEN_ESCORTING)  
      
      escort_ap->dir = aliens[fs].dir;
    }
    else
    {
      escort_ap->path = -1;
      escort_ap->steer = 2;
      escort_ap->escorting = -1;
      escort_ap->state=ATTACKING;
    }
  }

static int alien_can_attack(alien_t* alien_ap)
  {
    /*If it's a gaurd, don't let it attack on its own while flagships are still alive.*/
    if(alien_ap->type == GAURD_SHIP_1 || alien_ap->type == GAURD_SHIP_12)
    {
      int e_i;
      for(e_i= 0;e_i < FLAGSHIPS;++e_i)
      {
        if(aliens[e_i].alive)
        {
          return 0;
        }
      }
    }
    return 
    (
      (entering == 0) &&
      (attacking < maxattacking) && 
      (
        (livecount < maxattacking) || 
        ((random()%10000) < (2+level+2*(killed_aliens)))
    ) 
  ); 
  }

static void
  alien_in_convoy(alien_t* convoy_ap)
  {
    convoy_ap->x += convoymove;
    if(alien_can_attack(convoy_ap))
    /*
      (
      (entering == 0) &&
      (attacking < maxattacking) && 
      (
      (livecount < maxattacking) || 
      ((random()%10000) < (2+level+2*(killed_aliens)))
      ) && (convoy_ap->type != GAURD_SHIP_1 && convoy_ap->type != GAURD_SHIP_12)
      )
    */
    {
      switch(random()%2)
      {
        case 0:
          path_dir(P_PEELLEFT, 0, &convoy_ap->dir, &convoy_ap->steer);
          convoy_ap->path = P_PEELLEFT;
          break;
        case 1:
          path_dir(P_PEELRIGHT, 0, &convoy_ap->dir, &convoy_ap->steer);
          convoy_ap->path = P_PEELRIGHT;
          break;
      }
      convoy_ap->path_pos = 0;
      convoy_ap->state=ATTACKING;
      attacking++;
      if(convoy_ap->type==FLAG_SHIP_0||convoy_ap->type==FLAG_SHIP_11)
      { /* Flagship, grab escorts */
        int escorts_i=0;
        int ei=0;
        for(ei=FLAGSHIPS;ei<MAXALIENS;++ei)
        {
          if(aliens[ei].alive)
          {
            if(aliens[ei].state==IN_CONVOY)
            {
              aliens[ei].escorting = convoy_ap->id;
              ++escorts_i;
              aliens[ei].state=ESCORTING;
              if(escorts_i>=3){break;}
            }
          }
        }
      }
    }
  }

static void new_etorp(double x, double y, double xs, double ys)
  {
    struct torp *t;

    t = malloc(sizeof(struct torp));
    t->next = first_etorp;
    if(t->next)
    t->next->prev = t;
    t->prev = 0;
    first_etorp = t;

    t->x = x;
    t->y = y;
    t->xspeed = xs;
    t->yspeed = ys;
    t->alive = 1;
    t->frame = 0;
    numetorps++;
  }

void maybe_throw_etorp(const alien_t* alien_ap,const player_t* player_pp,int aim_level_i)
  {
    int tc=0;
    tc=TORPCHANCE-level/2-player_pp->weapon*5;
    if(tc<WORST_TORPCHANCE){tc=WORST_TORPCHANCE;}
    if(numetorps<maxetorps&&(!(random()%tc)))
    {
      double miss_p_x_d=0;
      double miss_n_x_d=0;
      double miss_p_y_d=0;
      double miss_n_y_d=0;
    
      double aim_x_d=0;
      double aim_y_d=0;
      double xspeed_d=0;
      double yspeed_d=0;
      double mag_d=0;

      double etorp_speed_d=0;
    
      int etorp_miss_i=(ETORP_MISS-(alien_ap->type<<1))-(aim_level_i<<1);
    
      if(etorp_miss_i<playerShip->height*2){etorp_miss_i=playerShip->height*2;}
      if(etorp_miss_i<playerShip->width*2){etorp_miss_i=playerShip->width*2;}

      miss_p_x_d=random()%etorp_miss_i;
      miss_n_x_d=-(random()%etorp_miss_i);
      miss_p_y_d=random()%etorp_miss_i;
      miss_n_y_d=-(random()%etorp_miss_i);

      aim_x_d=miss_p_x_d+miss_n_x_d+player_pp->x;
      aim_y_d=miss_p_y_d+miss_n_y_d+player_pp->y;
      aim_x_d-=alien_ap->x;
      aim_y_d-=alien_ap->y;
      mag_d=(sqrt(aim_x_d*aim_x_d+aim_y_d*aim_y_d));
      etorp_speed_d=(ETORPSPEED+aim_level_i/5);
      xspeed_d=(aim_x_d/mag_d)*(etorp_speed_d);
      yspeed_d=(aim_y_d/mag_d)*(etorp_speed_d);
      new_etorp(alien_ap->x,alien_ap->y,xspeed_d,yspeed_d);
    }
  }

static void alien_waiting_to_enter(alien_t* waiting_ap)
  {
    waiting_ap->enterdelay--;
#if defined(DEBUG_DO_ENTER)
    fprintf(OUT_FILE,"Alien %i delay %i before entry.\n",waiting_ap->id,waiting_ap->enterdelay);
#endif /* defined(DEBUG_DO_ENTER)        */
    if(waiting_ap->enterdelay<0)
    {
#if defined(DEBUG_DO_ENTER)
      fprintf(OUT_FILE,"Alien %i changing to ENTERING state.\n",waiting_ap->id);
#endif /* defined(DEBUG_DO_ENTER)        */
      waiting_ap->state=ENTERING;
    }
  }

static void alien_enter(alien_t* entering_ap)
  {
    if(entering_ap->path >= 0)
    {
#if defined(DEBUG_DO_ENTER)
      fprintf
      (
        OUT_FILE,
        "Alien %i entering, now at (%5.1f,%5.1f) .\n",entering_ap->id,entering_ap->x,entering_ap->y
    );
#endif /* defined(DEBUG_DO_ENTER)        */
    
      entering_ap->steer--;
      if(!entering_ap->steer)
      {
        entering_ap->path_pos++;
        path_dir(entering_ap->path, entering_ap->path_pos, &entering_ap->dir, &entering_ap->steer);
#if defined(DEBUG_DIR)
        if(entering_ap->dir<0||entering_ap->dir>15)
        {
          fprintf
          (
            OUT_FILE,
            "Alien %i with path %i and path_pos %i got invalid dir %i !\n",
            entering_ap->id,entering_ap->path,entering_ap->path_pos,entering_ap->dir
        );
          fprintf
          (
            OUT_FILE,"  Check array paths and function path_dir() in file paths.c for errors."
        );
        }
#endif /* defined(DEBUG_DIR)*/
        if(entering_ap->steer < 0)
        {
#if defined(DEBUG_DO_ENTER)
          fprintf
          (
            OUT_FILE,
            "Alien %i done entering, changing to RETURNING_TO_CONVOY state.\n",entering_ap->state
        );
#endif /* defined(DEBUG_DO_ENTER)        */
          entering_ap->state=RETURNING_TO_CONVOY;
        }
        entering_ap->steer = 2*(entering_ap->steer/3);
      }
    }
  }

void collision_check_player(struct alien* crunch_ap,struct player* player_pp)
  {
#if defined(DEBUG_COLLISION_CHECK_PLAYER)
    fprintf
    (
      OUT_FILE,"Checking alien at (%5.1f,%5.1f) against player at (%5.1f,%5.1f).\n",
      crunch_ap->x,crunch_ap->y,player_pp->x,player_pp->y
    );
#endif /*defined(DEBUG_COLLISION_CHECK_PLAYER)  */
    if(crunch_ap->state<=WAITING_TO_ENTER){return;}
    if
    (
      (!player_pp->flash)&&
      (!player_pp->dead)&&
      (crunch_ap->alive)&&
      (
        (crunch_ap->x)>(player_pp->x-(playerShip->width/2))&&
        (crunch_ap->x)<(player_pp->x+(playerShip->width/2))
      )&&
      (
        (crunch_ap->y)>(player_pp->y-(playerShip->height/2))&&
        (crunch_ap->y)<(player_pp->y+(playerShip->height/2))
      )
    )
    {
#if defined(DEBUG_COLLISION_CHECK_PLAYER)
      fprintf
      (
        OUT_FILE,"Alien at (%5.1f,%5.1f) hit player at (%5.1f,%5.1f).\n",
        crunch_ap->x,crunch_ap->y,player_pp->x,player_pp->y
      );
#endif /*defined(DEBUG_COLLISION_CHECK_PLAYER)  */
      crunch_ap->dying=1;
      if(player_pp->shield)
      {
        torp_hit_alien(crunch_ap);
      }
      else if(player_0.safety_bubbles>0)
      {
        torp_hit_alien(crunch_ap);
        --player_0.safety_bubbles;
      }
      else
      {
        player_pp->dead=1;
      }
      new_explosion(player_pp->x, player_pp->y, 2);
    }
  }

static void alien_moving(alien_t* mover_ap)
  {
    if(mover_ap->state==WAITING_TO_ENTER||mover_ap->state==IN_CONVOY)
    {
      return;
    }
    maybe_throw_etorp(mover_ap,&player_0,level);
    mover_ap->x += moves[mover_ap->dir][0]*1.5;
    mover_ap->y += moves[mover_ap->dir][1]*1.5;
    /*
      Since most of the aliens *start* out of bounds, don't do bounds
      checking on aliens that are still entering.
    */
    if(mover_ap->state==ENTERING)
    {
      return;
    }
    if(mover_ap->x > SPACE_X_MAX)
    {
      mover_ap->x = SPACE_X_MIN;
    }
    else if(mover_ap->x < SPACE_X_MIN)
    {
      mover_ap->x = SPACE_X_MAX;
    }
    if(mover_ap->y<SPACE_Y_MIN)
    {
      mover_ap->y=SPACE_Y_MAX;
    }
    else if(mover_ap->y > SPACE_Y_MAX)
    {
      mover_ap->y = SPACE_Y_MIN;
      mover_ap->path = -1;
      mover_ap->steer = 2;
      mover_ap->escorting = -1;
      attacking--;
      mover_ap->state=RETURNING_TO_CONVOY;
      if(mover_ap->type==FLAG_SHIP_0||mover_ap->type==FLAG_SHIP_11)
      {
        int j=0;
        for(j=FLAGSHIPS;j<FLAGSHIPS_AND_GAURDS;j++)
        {
          if(aliens[j].escorting==mover_ap->id)
          {
            aliens[j].escorting = -1;
            aliens[j].state=RETURNING_TO_CONVOY;
          }
        }
      }
    }
    else if(mover_ap->y < 0)
    {
      mover_ap->dir = 8;
    }
  }

static void alien_attacking(alien_t* attacker_ap)
  {
    attacker_ap->steer--;
    if(attacker_ap->steer <= 0)
    {
      if(attacker_ap->path >= 0)
      {
#if defined(DEBUG_DIR)
        int lastdir=attacker_ap->dir;
#endif /*defined(DEBUG_DIR)        */
                
        attacker_ap->path_pos++;
        path_dir(attacker_ap->path, attacker_ap->path_pos, &attacker_ap->dir, &attacker_ap->steer);

        /* Has the alien reached the end of its path?*/
        if(attacker_ap->steer < 0)
        {
#if defined(DEBUG_DIR)
          if(attacker_ap->dir<0||attacker_ap->dir>15)
          {
            fprintf
            (
              OUT_FILE,"Alien %i has invalid dir %i ! Resetting dir to last dir.\n",
              attacker_ap->id,attacker_ap->dir
          );
            attacker_ap->dir=lastdir;
          }
#endif /*defined(DEBUG_DIR)        */

          do
          {
            switch(random()%8)
            {
              case 0:
                start_path(P_LOOP, attacker_ap);
                break;
              case 1:
                start_path(P_SWOOP1, attacker_ap);
                break;
              case 2:
                start_path(P_SWOOP2, attacker_ap);
                break;
              case 3:
                start_path(P_ZIGZAG, attacker_ap);
                break;
              case 4:
                start_path(P_LOOP2, attacker_ap);
                break;
              case 5:
                start_path(P_SPIN, attacker_ap);
                break;
              case 6:
                start_path(P_LEFTDIAG, attacker_ap);
                break;
              case 7:
                start_path(P_RIGHTDIAG, attacker_ap);
                break;
              default:
                attacker_ap->steer = TURNSPEED;
                attacker_ap->path = -1;
            }
          }while((attacker_ap->path < 0) || (attacker_ap->steer < 0)) ;
        }
      }
      else
      {
        if(random()%2)
        {
          attacker_ap->dir++;
          if(attacker_ap->dir > 15)
          {
            attacker_ap->dir = 0;
          }
        }
        else
        {
          attacker_ap->dir--;
          if(attacker_ap->dir < 0)
          {
            attacker_ap->dir = 15;
          }
        }
        attacker_ap->steer = TURNSPEED;
      }
    }
  }

static void alien_returning_to_convoy(alien_t* returner_ap)
  {
    int diffx=0;
    int diffy=0;
    diffx = ABS(convoy_x_pos(returner_ap->id) - returner_ap->x);
    diffy = ABS(convoy_y_pos(returner_ap->id) - returner_ap->y);
    if(diffy<5)
    {
      returner_ap->y = convoy_y_pos(returner_ap->id);
      if(diffx < 5)
      {
        returner_ap->x = convoy_x_pos(returner_ap->id);
        returner_ap->entering = 0;
        returner_ap->state=IN_CONVOY;
        returner_ap->dir=0;
        return;
      }
      returner_ap->dir = 0;
      if(convoy_x_pos(returner_ap->id) > returner_ap->x)
      {
        returner_ap->dir = 4;
      }
      else
      {
        returner_ap->dir = 12;
      }
    }
    else
    {
      if(diffx < 5)
      {
        returner_ap->x = convoy_x_pos(returner_ap->id);
        if(returner_ap->y<convoy_y_pos(returner_ap->id))
        {
          returner_ap->dir = 8;
        }
        else
        {
          returner_ap->dir = 0;
        }
      }
      else
      {
        if(convoy_x_pos(returner_ap->id) > returner_ap->x)
        {
          if(returner_ap->y<convoy_y_pos(returner_ap->id))
          {
            returner_ap->dir = 10;
          }
          else
          {
            returner_ap->dir = 14;
          }
        }
        else
        {
          if(returner_ap->y<convoy_y_pos(returner_ap->id))
          {
            returner_ap->dir = 6;
          }
          else
          {
            returner_ap->dir = 2;
          }
        }
      }
    }
#if defined(DEBUG_DO_ENTER)
    fprintf
    (
      OUT_FILE,
      "Alien %i moving into convoy, now at (%5.1f,%5.1f) .\n",returner_ap->id,
      returner_ap->x,returner_ap->y
  );
#endif /* defined(DEBUG_DO_ENTER)        */
    if
    (
      (returner_ap->x<-4*WINWIDTH||returner_ap->x>5*WINWIDTH)||
      (returner_ap->y<-4*WINHEIGHT||returner_ap->y>5*WINHEIGHT)
  )
    {
      fprintf(OUT_FILE,"  Alien %i going to hell!\n",returner_ap->id);
      fprintf(OUT_FILE,"  Killing Alien %i at position (%5.1f,%5.1f).\n",returner_ap->id,returner_ap->x,returner_ap->y);
      returner_ap->alive=0;
    }
  }

static void do_aliens()
  {
    int i;

    /*
      Move the convoy.
    */
    if(!paused)
    {
      convoyx += convoymove;
      if(convoyx <= 0)
      {
        convoyx=0;
        convoymove = -convoymove;
      }
      else if(convoyx >= WINWIDTH/2)
      {
        convoyx = WINWIDTH/2;
        convoymove = -convoymove;
      }
    }
    
    livecount=0; attacking = 0;
    for(i=0, livecount=0, entering=0; i < MAXALIENS; i++)
    {
      if(aliens[i].alive)
      {
        livecount++;
        if(aliens[i].state==ATTACKING)
        {
          attacking++;
        }
        if(aliens[i].state==ENTERING)
        {
          entering++;
        }
      }
    }
#if defined(DEBUG_DO_ALIENS)
    fprintf(OUT_FILE,"%i living aliens.",livecount);
#endif /* defined(DEBUG_DO_ALIENS)        */

    for(i=0; i < MAXALIENS; i++)
    {
      if(aliens[i].alive)
      {
        if(!paused)
        {
          collision_check_player(&aliens[i],&player_0);

          alien_moving(aliens+i);
                
          switch(aliens[i].state)
          {
            case WAITING_TO_ENTER:
              alien_waiting_to_enter(aliens+i);
              break;
            case ENTERING:
              alien_enter(aliens+i);
              break;
            case IN_CONVOY:
              alien_in_convoy(aliens+i);
              break;
            case ATTACKING:
              alien_attacking(aliens+i);
              break;
            case ESCORTING:
              alien_escorting(aliens+i);
              break;
            case RETURNING_TO_CONVOY:
              alien_returning_to_convoy(aliens+i);
              break;
            case DEAD:
              /* It's dead. What do you expect me to do? Resurrect it?*/
              break;
            default:
              fprintf(OUT_FILE,"Alien %i in unknown state %i!\n",i,aliens[i].state);
          }
#if defined(DEBUG_DIR)
          if((aliens[i].dir<0)||(aliens[i].dir>15))
          {
            fprintf(OUT_FILE,"Alien %i with invalid dir %i! Resetting to 0.",aliens[i].id,aliens[i].dir);
          }
#endif /* defined(DEBUG_DIR)*/


          W_DrawImage
          (
            baseWin, 
            (int)aliens[i].x-(aliens[i].shape->width/2), 
            (int)aliens[i].y-(aliens[i].shape->height/2),
            aliens[i].dir,
            aliens[i].shape, W_Green
        );
        }
        else
        {  /* paused */
          W_DrawImage
          (
            baseWin, 
            (int)aliens[i].x-(aliens[i].shape->width/2), 
            (int)aliens[i].y-(aliens[i].shape->height/2),
            aliens[i].dir,
            aliens[i].shape, W_Green
        );
        }
      }
    }
    if(livecount == 0)
    {
      starspeed++;
#ifdef SOUND
      if(starspeed == 2)
      {
        play_sound(SND_WARP);
      }
#endif
      if(starspeed >= 120) {
        starspeed = -20;
      }
      else if(starspeed == 1)
      {
        init_aliens(++level);
        player_0.got_lemon = 0;
        starspeed = 1;
        player_0.numtorps=0;
      }
    }
  }


/*------------------player----------------*/
void reset_player(player_t* _player_pp,int level)
  {
    _player_pp->y= WINHEIGHT-playerShip->height/2;
    _player_pp->x= WINWIDTH/2;
    _player_pp->movespeed= MINSPEED;
    _player_pp->maxtorps= MINTORPS;
    _player_pp->weapon= SINGLESHOT;
    _player_pp->weapon+= (cheat_6_shot*level*2) + (cheat_7_shot*2*level*2);
    _player_pp->maxtorps+= _player_pp->weapon*2;
    _player_pp->shield= 0;
    _player_pp->got_lemon= 0;
    _player_pp->dead= 0;
    _player_pp->flash= FLASH_TIME;
    _player_pp->times_torps_can_wrap= 0;
    _player_pp->safety_bubbles= 0;
    _player_pp->cannon_power= 1;
    _player_pp->shoot_down_etorps= 0;
    _player_pp->torp_velocity= 0;
    _player_pp->prize_velocity= 0;
  }

static void init_player(player_t* _player_pp,int level)
  {
    int i;
    for(i=0;i<MAXTORPS;i++)
    {
      torps[i].alive=0;
    }
    reset_player(_player_pp,level);
    _player_pp->numtorps=0;
    _player_pp->next_bonus=START_NEXT_BONUS;
    _player_pp->ships=START_SHIPS+level/8;
    _player_pp->score=0;
  }

/*static void new_cluster(double x, double y, double xs, double ys)
  {
    int i;

      if(!cluster[i].alive)
      {
        clusters[i].x = x;
        clusters[i].y = y;
        clusters[i].alive = 1;
        clusters[i].xspeed = xs;
        clusters[i].yspeed = ys;

#ifdef SOUND
        play_sound(SND_FIRETORP);
#endif
        return;
      }
    }

static void undo_clusters()
  {
    int i;

      if(clusters[i].alive) {
        W_CacheClearArea(baseWin, 
          clusters[i].x-(playerCluster->width/2), 
          clusters[i].y-(playerCluster->height/2), 
          playerCluster->width, playerCluster->height+1);
      }
      }*/

/*static void do_clusters()
  {
    int i,j;

    if(clusters[i].alive > 0)
      {
        if(!paused)
        {
          int torp_wrapped= 0;
          clusters[i].y += clusters[i].yspeed;
          clusters[i].x += clusters[i].xspeed;
          clusters[i].frame++;
          for(j=0;j<MAXALIENS;j++)
          {
            if
            (
              (aliens[j].alive) && (!aliens[j].dying) &&
              (aliens[j].state!=WAITING_TO_ENTER)&&
              (ABS(clusters[i].x - aliens[j].x) < 8) &&
              (
                (ABS(clusters[i].y-aliens[j].y) < 8) ||
                (ABS((clusters[i].y+clusters[i].yspeed/2)-aliens[j].y) < 8)
              )
            )
            {
              torp_hit_alien(&aliens[j]);
              clusters[i].alive=0;
            }
          }
          if (clusters[i].alive)
          {
            W_DrawImage
            (
              baseWin, 
              (int)clusters[i].x-(playerCluster->width/2), 
              (int)clusters[i].y-(playerCluster->height/2), 
              clusters[i].frame, playerCluster, W_Red
            );
          }
        } else {paused
W_DrawImage
          (
            baseWin, 
            (int)clusters[i].x-(playerCluster->width/2), 
            (int)clusters[i].y-(playerCluster->height/2), 
            clusters[i].frame, playerCluster, W_Red
          );
        }
      }
      }
  }*/

static void new_torp(double x, double y, double xs, double ys)
  {
    int i;

    for(i=0;i<MAXTORPS;i++)
    {
      if(!torps[i].alive)
      {
        torps[i].x = x;
        torps[i].y = y;
        torps[i].alive = 1;
        torps[i].xspeed = xs;
        torps[i].yspeed = ys;
            
        torps[i].wrap_times_left= player_0.times_torps_can_wrap;
        player_0.numtorps++;
#ifdef SOUND
        play_sound(SND_FIRETORP);
#endif
        return;
      }
    }
  }

static void undo_torps()
  {
    int i;

    for(i=0;i<MAXTORPS;i++) {
      if(torps[i].alive) {
        W_CacheClearArea(baseWin, 
          torps[i].x-(playerTorp->width/2), 
          torps[i].y-(playerTorp->height/2), 
          playerTorp->width, playerTorp->height+1);
      }
    }
  }

static void do_torps()
  {
    int i,j;

    for(i=0;i<MAXTORPS;i++)
    {
      if(torps[i].alive > 0)
      {
        if(!paused)
        {
          int torp_wrapped= 0;
          torps[i].y += torps[i].yspeed;
          torps[i].x += torps[i].xspeed;
          torps[i].frame++;
          for(j=0;j<MAXALIENS;j++)
          {
            if
            (
              (aliens[j].alive) && (!aliens[j].dying) &&
              (aliens[j].state!=WAITING_TO_ENTER)&&
              (ABS(torps[i].x - aliens[j].x) < 8) &&
              (
                (ABS(torps[i].y-aliens[j].y) < 8) ||
                (ABS((torps[i].y+torps[i].yspeed/2)-aliens[j].y) < 8)
              )
            )
            {
              torp_hit_alien(&aliens[j]);
              torps[i].alive=0;
              --player_0.numtorps;
            }
          }         
          if (torps[i].x>SPACE_X_MAX)
          {
            torp_wrapped= 1;
            torps[i].x= SPACE_X_MIN;
          }
          if (torps[i].y>SPACE_Y_MAX)
          {
            torp_wrapped= 1;
            torps[i].y= SPACE_Y_MIN;
          }
          if (torps[i].x<SPACE_X_MIN)
          {
            torp_wrapped= 1;
            torps[i].x= SPACE_X_MAX;
          }
          if (torps[i].y<SPACE_Y_MIN)
          {
            torp_wrapped= 1;
            torps[i].y= SPACE_Y_MAX;
          }
          if (torp_wrapped)
          {
            if (torps[i].wrap_times_left > 0 && (player_0.maxtorps - player_0.numtorps) > player_0.weapon*2)
            {
              --torps[i].wrap_times_left;
            }
            else
            {
              torps[i].alive= 0;
              --player_0.numtorps;
            }
          }
          if (torps[i].alive)
          {
            W_DrawImage
            (
              baseWin, 
              (int)torps[i].x-(playerTorp->width/2), 
              (int)torps[i].y-(playerTorp->height/2), 
              torps[i].frame, playerTorp, W_Red
            );
          }
        } else {/* paused */
          W_DrawImage
          (
            baseWin, 
            (int)torps[i].x-(playerTorp->width/2), 
            (int)torps[i].y-(playerTorp->height/2), 
            torps[i].frame, playerTorp, W_Red
          );
        }
      }
    }
  }
void torp_hit_alien(alien_t* victim_ap)
  {
    score_alien_hit(victim_ap);
    if(victim_ap->alive>1)
    {
      --(victim_ap->alive);
      return;
    }
    ++killed_aliens;
    victim_ap->dying=1;
    switch(victim_ap->state)
    {
      case WAITING_TO_ENTER:
        fprintf
        (
          OUT_FILE,
          "Error! torp_hit_alien() called on alien %i in state WAITING_TO_ENTER !\n",
          victim_ap->id
        );
        break;
      case ENTERING:
        --entering;
        break;
      case IN_CONVOY:
        /* Do nothing.*/
        break;
      case ATTACKING:
        --attacking;
        break;
      case ESCORTING:
        /* Do nothing. */
        break;
      case RETURNING_TO_CONVOY:
        /* Do nothing. */
        break;
      case DEAD:
        fprintf
        (
          OUT_FILE,
          "Error! torp_hit_alien() called on alien %i in state DEAD !\n",
          victim_ap->id
        );
        break;
      default:
        fprintf
        (
          OUT_FILE,
          "Error! torp_hit_alien() called on alien %i in unknown state !\n",
          victim_ap->id
        );
    }
    victim_ap->state=DEAD;
#if defined(DEBUG_DIR)
    if(victim_ap->dir<0||victim_ap->dir>15)
    {
      fprintf(OUT_FILE,"Alien %i with invalid dir %i! Resetting to 0.",victim_ap->id,victim_ap->dir);
    }
#endif defined(DEBUG_DIR)  
  }

void score_alien_hit(const alien_t* victim_ap)
  {
    double prize_chance= PRIZECHANCE;
    int score_alien= 0;
    switch(victim_ap->type)
    {
      case FLAG_SHIP_0:
      case FLAG_SHIP_11:
        score_flag_ship_hit(victim_ap);
        new_explosion(victim_ap->x,victim_ap->y,1);
        break;
      default:
        new_explosion(victim_ap->x,victim_ap->y,0);
        {
          int mul_i=10;
          switch(victim_ap->state)
          {
            case WAITING_TO_ENTER:
              fprintf
              (
                OUT_FILE,
                "Error! score_alien_hit() called on alien %i in state WAITING_TO_ENTER !\n",
                victim_ap->id
              );
              break;
            case ENTERING:
              mul_i<<=1;
              break;
            case IN_CONVOY:
              /* Do nothing.*/
              break;
            case ATTACKING:
              mul_i<<=2;
              break;
            case ESCORTING:
              mul_i<<=2;
              break;
            case RETURNING_TO_CONVOY:
              mul_i<<=1;
              break;
            case DEAD:
              fprintf
              (
                OUT_FILE,
                "Error! score_alien_hit() called on alien %i in state DEAD !\n",
                victim_ap->id
              );
              break;
            default:
              fprintf
              (
                OUT_FILE,
                "Error! score_alien_hit() called on alien %i in unknown state !\n",
                victim_ap->id
              );
          }
          score_alien= mul_i*victim_ap->type;
          player_0.score+= score_alien;
          prize_chance= PRIZECHANCE - (PRIZECHANCE - ((PRIZECHANCE)/((score_alien/40.0) + level)))*0.8;
          if (player_0.got_lemon)
          {
            prize_chance/= 4;
          }
          /*
          fprintf(OUT_FILE,"In %s prize_chance == %f, type == %i, level == %i.\n",
            __PRETTY_FUNCTION__, prize_chance, victim_ap->type, level);
          */
          if(!(random()%((int)prize_chance)))
          {
            new_prize(victim_ap->x,victim_ap->y);
          }
        }
    }
  }

void score_flag_ship_hit(const alien_t* victim_ap)
  {
    double prize_chance= PRIZECHANCE;
    int escorts_i=0;
    int ki=0;

    for(ki=FLAGSHIPS;ki<MAXALIENS;++ki)
    {
      if(aliens[ki].escorting==victim_ap->id)
      {
        ++escorts_i;
      }
    }
    prize_chance= PRIZECHANCE - (PRIZECHANCE - ((PRIZECHANCE)/(victim_ap->type*2.0 + level + escorts_i)))*0.8;
    if (player_0.got_lemon)
    {
      prize_chance/= 4;
    }
    /*
    fprintf(OUT_FILE,"In %s prize_chance == %f, type == %i, level == %i.\n",
      __PRETTY_FUNCTION__, prize_chance, victim_ap->type, level);
    */
    if(!(random()%((int)prize_chance)))
    {
      new_prize(victim_ap->x,victim_ap->y);
    }
    score_flagship(victim_ap->x,victim_ap->y,escorts_i);
  }


static void do_etorps()
  {
    struct torp *t = first_etorp, *nextt;
    int i;

    i= 0;
    
    while(t) {
      nextt=t->next;
      if(t->alive) {
        if(!paused) {
          t->y+=t->yspeed;
          t->x+=t->xspeed;
          t->frame++;
          if(player_0.shoot_down_etorps>0)
          {
            for (i= 0 ; i < MAXTORPS ; ++i)
            {
              /*Only collison check player torps that are alive.*/
              if(torps[i].alive > 0)
              {
                if((ABS(t->x - torps[i].x) < 4) && (ABS(t->y - torps[i].y) < 4))
                {
                  --t->alive;
                  --torps[i].alive;
                  --player_0.numtorps;
                  W_DrawImage
                  (
                    baseWin, 
                    (int)t->x-(enemyTorp->width/2), 
                    (int)t->y-(enemyTorp->height/2), 
                    t->frame, enemyTorp, W_Red
                    );
                }
              }
            }
          }
          if((t->y>SPACE_Y_MAX)||(t->y<SPACE_Y_MIN)||(t->x<SPACE_X_MIN)||(t->x>SPACE_X_MAX))
          {
            t->alive= 0;
          } else if(!player_0.dead && !player_0.flash && !player_0.shield &&
            (ABS(t->x - player_0.x) < 8) &&
            (ABS(t->y - player_0.y) < 8))
          { /* DEAD! */
            /*
              Check for safety bubbles.
              If the player has one or more safety bubbles,
              decrement the number of safety bubbles, and kill the
              etorp, but do
              not kill the player.
            */
            if(player_0.safety_bubbles>0)
            { /* safe! */
              --player_0.safety_bubbles;
              t->alive= 0;
            }
            else
            {
              player_0.dead = 1;
              new_explosion(player_0.x, player_0.y, 2);
              t->alive= 0;
              player_0.dead = 1;
              new_explosion(player_0.x, player_0.y, 2);
              t->alive= 0;
            }
          } else {
            W_DrawImage
            (
              baseWin, 
              (int)t->x-(enemyTorp->width/2), 
              (int)t->y-(enemyTorp->height/2), 
              t->frame, enemyTorp, W_Red
            );
          }
        } else {
          W_DrawImage
          (
            baseWin, 
            (int)t->x-(enemyTorp->width/2), 
            (int)t->y-(enemyTorp->height/2), 
            t->frame, enemyTorp, W_Red);
        }
      }
      t = nextt;
    }
  }

static void undo_etorps()
  {
    torp_t *t = first_etorp;
    torp_t *nextt;

    while(t)
    {
      nextt= t->next;
      W_CacheClearArea
      (
        baseWin, 
        (int)t->x-(enemyTorp->width/2), 
        (int)t->y-(enemyTorp->width/2), 
        enemyTorp->width, enemyTorp->height
      );
      if (!t->alive)
      {
        if(t->next)
        {
          t->next->prev = t->prev;
        }
        if(t->prev)
        {
          t->prev->next = t->next;
        }
        if(t == first_etorp)
        {
          first_etorp = t->next;
        }
        free(t);
        if((!player_0.dead)&&(!player_0.shield)&&(!player_0.flash))
        {
          ++player_0.score;
        }
        numetorps--;
      }
      t=nextt;
    }
  }

static void undo_player()
  {
    int h = playerShip->height;
    int y = player_0.y-h/2;
    if(player_0.shield || player_0.safety_bubbles) {
      y -= 3;
      h += 3;
    }
    W_CacheClearArea
    (
      baseWin,
      (int)player_0.x-(playerShip->width/2), y,
      playerShip->width, h
    );
  }

int get_torps_to_launch(int torps_in_use, int max_torps, weapon_type_t weapon)
  {
    int torps_remaining= max_torps - torps_in_use;
    ++weapon;/*SINGLESHOT is weapon 0 - so add 1 to allow firing of at least 1 shot.*/
    if (torps_remaining > 0)
    {
      if (weapon > torps_remaining)
      {
        return torps_remaining;
      }
      return weapon;
    }
    return 0;
  }

void spread_torps(player_t* player_p, double start_angle, double inc_angle, int torps)
  {
    int current;
    double base_torp_speed= TORPSPEED+player_0.torp_velocity;
    double torp_x_speed;
    double torp_y_speed;
    double angle;
    for (current= 0, angle= start_angle; current < torps ; ++current, angle+= inc_angle)
    {
      torp_x_speed= cos(angle)*base_torp_speed;
      /* The y speed is negative because y increases going *down*. */
      torp_y_speed= -sin(angle)*base_torp_speed;
      //fprintf(OUT_FILE, "In %s, line %d, file %s : torp_x_speed == %f, torp_y_speed == %f\n", __PRETTY_FUNCTION__, __LINE__, __FILE__, torp_x_speed, torp_y_speed );
      new_torp( player_p->x + torp_x_speed, player_p->y + torp_y_speed, torp_x_speed, torp_y_speed);
    }
  }

void double_torp(double x, double y, double xs, double ys)
  {
    new_torp(x-ys/4, y+xs/4, xs, ys);
    new_torp(x+ys/4, y-xs/4, xs, ys);
  }

void fire_torps2(player_t* player_p)
  {
    int torps_to_launch;
    int torps_straight_up;
    int torps_spread_to_each_side;
    double base_torp_speed= TORPSPEED+player_0.torp_velocity;
    double angle;
    double torp_x_speed;
    double torp_y_speed;
    double spread_angle= M_PI/25.0*player_0.cannon_power;
    
    torps_to_launch= get_torps_to_launch(player_p->numtorps, player_p->maxtorps, player_p->weapon);
    //fprintf(OUT_FILE, "In %s, line %d, file %s : torps_to_launch == %d\n", __PRETTY_FUNCTION__, __LINE__, __FILE__, torps_to_launch );
    if (torps_to_launch <= 0)
    {
      return;
    }

    /*
      If we are launching an even number of torps, we want 2 torps to go straight up.
      If we are launching an odd number of torps, we want 1 torp to go straight up.
      % == remainder operator. Divide by 2 and return the remainder.
     */
    if (torps_to_launch%2 == 0) 
    {
      torps_straight_up= 2; /*even*/
    }
    else
    {
      torps_straight_up= 1; /*odd*/
    }

    //fprintf(OUT_FILE, "In %s, line %d, file %s : torps_straight_up == %d\n", __PRETTY_FUNCTION__, __LINE__, __FILE__, torps_straight_up );
    /*
      Launch the torps that go straight up.
     */
    angle= M_PI/2.0; /* pi/2 radians is straight up.*/
    torp_x_speed= cos(angle)*base_torp_speed;
    /* The y speed is negative because y increases going *down*. */
    torp_y_speed= -sin(angle)*base_torp_speed;
    switch (torps_straight_up)
    {
      case 1:
        new_torp( player_p->x + torp_x_speed, player_p->y + torp_y_speed, torp_x_speed, torp_y_speed);
        break;
      case 2:
        double_torp( player_p->x + torp_x_speed, player_p->y + torp_y_speed, torp_x_speed, torp_y_speed);
        break;
    }

    torps_spread_to_each_side= (torps_to_launch - torps_straight_up)/2;

    //fprintf(OUT_FILE, "In %s, line %d, file %s : torps_spread_to_each_side == %d\n", __PRETTY_FUNCTION__, __LINE__, __FILE__, torps_spread_to_each_side );
    if (torps_spread_to_each_side <= 0)
    {
      return;
    }
    /*
      lauch the torps to the right.
      IF you need to change the spread_angle, do it before spread_torps is called.
     */
    spread_torps(player_p, angle+spread_angle, spread_angle, torps_spread_to_each_side);
    /*
      lauch the torps to the left.
     */
    spread_torps(player_p, angle-spread_angle, -spread_angle, torps_spread_to_each_side);
  }

/*void fire_cluster(player_p)
  {
  double spread_angle=*/
static void do_player()
{
    int mx, my, but;
    static int torpok;
    W_Event wev;
    int l=1;

    if(gameOver) {
      cheat_7_shot= 0;
      cheat_6_shot= 0;
        while(W_EventsPending()) {
            W_NextEvent(&wev);
        
            player_0.mouse_control = 1;
            switch(wev.type) {
            case W_EV_KEY:
                if(score_key(&wev))
                    continue;
                switch(wev.key)
                {
                  case 'q':
                  case 'Q':
                    xgal_exit(0);
                    break;
                  case 'K':
                  case 'k':
                    l=1;
                    player_0.mouse_control = 0;
                    break;
                  case '1':
                  case '2':
                  case '3':
                  case '4':
                  case '5':
                  case '6':
                  case '7':
                  case '8':
                  case '9':
                    l= wev.key-'0';
                    break;
                  case '0':
                    l=10;
                    break;
                  case '!':
                    l=12;
                    break;
                  case '@':
                    l=14;
                    break;
                  case '#':
                    l=16;
                    break;
                  case '$':
                    l=18;
                    break;
                  case '%':
                    l=20;
                    break;
                  case '^':
                    l=22;
                    break;
                  case '&':
                    l=24;
                    break;
                  case '*':
                    l=26;
                    break;
                  case '(':
                    l=28;
                    cheat_6_shot= 1;
                    break;
                  case ')':
                    l=30;
                    cheat_7_shot= 1;
                    break;
                }
                if(player_0.mouse_control < 2)
                    player_0.mouse_control = 0;
            case W_EV_BUTTON:
                if(!getting_name) {
                    if(player_0.mouse_control)
                        W_GrabPointer(baseWin);
                    gameOver = 0;
                    level=l;
                    init_aliens(level);
                    init_player(&player_0,level);
                    if (cheat_6_shot)
                    {
                      player_0.weapon= 25;
                      player_0.maxtorps= 30;
                      player_0.cannon_power= 3;
                      player_0.score-= player_0.weapon*10000;
                    }
                    if (cheat_7_shot)
                    {
                      player_0.weapon= 52;
                      player_0.maxtorps= 40;
                      player_0.cannon_power= 3;
                      player_0.score-= player_0.weapon*10000;
                    }
                    W_ClearWindow(baseWin);
                }
                break;
            }
        }
        return;
    }

    while(W_EventsPending()) {
        W_NextEvent(&wev);

        switch(wev.type)
        {
        case W_EV_KEY_OFF:
          switch(wev.key)
          {
            case RIGHT_LETTER:
            case RIGHT_ARROW:
            case RIGHT_NUM:
              player_0.keys &= ~(RIGHTKEY);
              break;
            case LEFT_LETTER:
            case LEFT_ARROW:
            case LEFT_NUM:
              player_0.keys &= ~(LEFTKEY);
              break;
            case UP_LETTER:
            case UP_NUM:
              player_0.keys &= ~(UPKEY);
              break;
            case DOWN_LETTER:
            case DOWN_NUM:
              player_0.keys &= ~(DOWNKEY);
              break;
            case ' ':
              player_0.keys &= ~(FIREKEY);
              break;
/*            case CLUSTER_NUM:
            case CLUSTER_LETTER:
              player_0.keys &= ~(CLUSTERKEY);
              break;*/
          }
          break;
          case W_EV_KEY:
#if defined(SHOW_KEY_DOWN)
          fprintf(OUT_FILE,"Key down: %i\n",wev.key);
#endif /* defined(SHOW_KEY_DOWN)          */
          switch(wev.key)
          {
            case RIGHT_LETTER:
            case RIGHT_ARROW:
            case RIGHT_NUM:
              player_0.keys |= RIGHTKEY;
              break;
            case LEFT_LETTER:
            case LEFT_ARROW:
            case LEFT_NUM:
              player_0.keys |= LEFTKEY;
              break;
            case UP_LETTER:
            case UP_NUM:
              player_0.keys |= UPKEY;
              break;
            case DOWN_LETTER:
            case DOWN_NUM:
              player_0.keys |= DOWNKEY;
              break;
            case ' ':
              player_0.keys |= FIREKEY;
              break;
/*            case CLUSTER_NUM:
            case CLUSTER_LETTER:
              player_0.keys |=CLUSTERKEY;
              break;*/
            case 'k':
              player_0.mouse_control = 0;
              W_UngrabPointer();
              break;
            case 'm':
              player_0.mouse_control = 1;
              W_GrabPointer(baseWin);
              break;
            case 'q':
              if(!player_0.dead) {
                new_explosion(player_0.x,player_0.y, 2);
                player_0.ships = 0;
                player_0.dead = 1;
              }
              break;
            case 'Q':
              xgal_exit(0);
              break;
            case 'b':
              W_Buffer(baseWin, !W_IsBuffered(baseWin));
              W_ClearWindow(baseWin);
              break;
            case 'p':
              paused=!paused;
              if(!paused)
              {
                undo_pause();
                if(player_0.mouse_control)
                W_GrabPointer(baseWin);
              } else
              W_UngrabPointer();
              break;
#ifdef SOUND
            case 's':
              playSounds = !playSounds;
              break;
#endif
#ifdef IM_A_BIG_FAT_CHEATER
            case 'i':
              player_0.score-=500;
              if(player_0.flash >= 0)
              player_0.flash = -2;
              else
              player_0.flash = 0;
              break;
            case 'l':
              player_0.score-=20000;
              {
                int i;
                for(i=0;i<MAXALIENS;i++)
                aliens[i].alive=0;
                if(starspeed != 1)
                level++;
              }
              break;
            case 'c':
              player_0.ships++;
              player_0.score-= 2*BONUSSHIPSCORE;
              break;
            case 'h':
              player_0.score-=5000;
              player_0.shield = SHIELDTIME;
#ifdef SOUND
              play_sound(SND_SHIELD);
#endif
              break;
            case 'w':
              player_0.score-=player_0.weapon*2000+2000;
              player_0.weapon++;
              if(player_0.weapon == 3)
              player_0.weapon=0;
              break;
            case 't':
              player_0.score-=player_0.maxtorps*1000;
              player_0.maxtorps++;
              if(player_0.maxtorps > MAXTORPS)
              player_0.maxtorps = MINTORPS;
              break;
#endif /* IM_A_BIG_FAT_CHEATER */
            default:
              break;
          }
        }
    }

    if(!paused) {
        torpok--;

        if(player_0.mouse_control) {
            W_GetMouse(baseWin, &mx, &my, &but);
        } else {
            if(player_0.keys & LEFTKEY)
                mx = 0;
            else if(player_0.keys & RIGHTKEY)
                mx = WINWIDTH;
            else
                mx = player_0.x;
        
            if(player_0.keys & FIREKEY)
                but = W_LBUTTON;
            else
                but = 0;
        }
        
        if(player_0.dead) {
            player_0.dead++;
            if
            (
              (player_0.keys&&player_0.dead>MIN_DEAD_TIME)
            )
            {
              player_0.keys=0;
                if(player_0.ships<=0) {
                    gameOver = 1;
                    W_UngrabPointer();
                    if(check_score(player_0.score)) {
#ifdef USE_REAL_NAMES
                        add_score(getUsersFullName(), player_0.score);
                        title_page = 1; pagetimer = 300;
#else
                        getting_name = 1;
#endif
                    }
                } else {
                    player_0.ships--;
                    reset_player(&player_0,level);
#if defined(PAUSE_ON_DEATH)                    
                    paused=1;
#endif /* defined(PAUSE_ON_DEATH)                    */
                }
            }
            return;
        }

        if(but && torpok <= 0 && (starspeed == 1))
        {
#if defined(ORIGINAL_FIRE)          
            switch(player_0.weapon)
            {
            case Q7_SHOT:
                if(player_0.numtorps < player_0.maxtorps-6) {
                    new_torp(player_0.x-15, player_0.y-(playerShip->height/2), -6, 1-TORPSPEED);
                    new_torp(player_0.x-10, player_0.y-(playerShip->height/2), -4, 1-TORPSPEED);
                    new_torp(player_0.x-5 , player_0.y-(playerShip->height/2), -2, 1-TORPSPEED);
                    new_torp(player_0.x   , player_0.y-(playerShip->height/2),  0, 1-TORPSPEED);
                    new_torp(player_0.x+5 , player_0.y-(playerShip->height/2),  2, 1-TORPSPEED);
                    new_torp(player_0.x+10, player_0.y-(playerShip->height/2),  4, 1-TORPSPEED);
                    new_torp(player_0.x+15, player_0.y-(playerShip->height/2),  6, 1-TORPSPEED);
                    break;
                }
            case Q6_SHOT:
                if(player_0.numtorps < player_0.maxtorps-5) {
                    new_torp(player_0.x-10, player_0.y-(playerShip->height/2), -4, 1-TORPSPEED);
                    new_torp(player_0.x-5 , player_0.y-(playerShip->height/2), -2, 1-TORPSPEED);
                    new_torp(player_0.x-5 , player_0.y-(playerShip->height/2),  0,  -TORPSPEED);
                    new_torp(player_0.x+5 , player_0.y-(playerShip->height/2),  0,  -TORPSPEED);
                    new_torp(player_0.x+5 , player_0.y-(playerShip->height/2),  2, 1-TORPSPEED);
                    new_torp(player_0.x+10, player_0.y-(playerShip->height/2),  4, 1-TORPSPEED);
                    break;
                }
            case Q5_SHOT:
                if(player_0.numtorps < player_0.maxtorps-4) {
                    new_torp(player_0.x-10, player_0.y-(playerShip->height/2), -4, 1-TORPSPEED);
                    new_torp(player_0.x-5 , player_0.y-(playerShip->height/2), -2, 1-TORPSPEED);
                    new_torp(player_0.x   , player_0.y-(playerShip->height/2),  0,  -TORPSPEED);
                    new_torp(player_0.x+5 , player_0.y-(playerShip->height/2),  2, 1-TORPSPEED);
                    new_torp(player_0.x+10, player_0.y-(playerShip->height/2),  4, 1-TORPSPEED);
                    break;
                }
            case QUAD_SHOT:
                if(player_0.numtorps < player_0.maxtorps-3) {
                    new_torp(player_0.x-10, player_0.y-(playerShip->height/2), -4, 1-TORPSPEED);
                    new_torp(player_0.x-5 , player_0.y-(playerShip->height/2),  0,  -TORPSPEED);
                    new_torp(player_0.x+5 , player_0.y-(playerShip->height/2),  0,  -TORPSPEED);
                    new_torp(player_0.x+10, player_0.y-(playerShip->height/2),  4, 1-TORPSPEED);
                    break;
                }
            case TRIPLESHOT:
                if(player_0.numtorps < player_0.maxtorps-2) {
                    new_torp(player_0.x-5, player_0.y-(playerShip->height/2), -2, 1-TORPSPEED);
                    new_torp(player_0.x,   player_0.y-(playerShip->height/2), 0,   -TORPSPEED);
                    new_torp(player_0.x+5, player_0.y-(playerShip->height/2), 2, 1-TORPSPEED);
                    break;
                }
            case DOUBLESHOT:
                if(player_0.numtorps < player_0.maxtorps-1) {
                    new_torp(player_0.x-5, player_0.y-(playerShip->height/2), 0, -TORPSPEED);
                    new_torp(player_0.x+5, player_0.y-(playerShip->height/2), 0, -TORPSPEED);
                    break;
                }
            case SINGLESHOT:
                if(player_0.numtorps < player_0.maxtorps)
                    new_torp(player_0.x, player_0.y-(playerShip->height/2), 0, -TORPSPEED);
                break;
            }
#else /* defined(ORIGINAL_FIRE)          */
            fire_torps2(&player_0);
            //fire_torps(&player_0,0,-TORPSPEED);
#endif /* defined(ORIGINAL_FIRE)          */
            torpok = TORPDELAY;
        }

        if(!but)
            torpok=0;

        if(player_0.mouse_control)
        {
          
          if((mx/player_0.movespeed) > (player_0.x/player_0.movespeed))
          {
            player_0.x+=(player_0.movespeed+level/5);
          }
          else if((mx/player_0.movespeed) < (player_0.x/player_0.movespeed))
          {
            player_0.x-=(player_0.movespeed+level/5);
          }
        }
        else
        {
          if(player_0.keys&LEFTKEY)
          {
            player_0.x-=(player_0.movespeed+level/5);
          }
          else if(player_0.keys&RIGHTKEY)
          {
            player_0.x+=(player_0.movespeed+level/5);
          }
          if(player_0.keys&UPKEY)
          {
            player_0.y-=(player_0.movespeed+level/5);
          }
          else if(player_0.keys&DOWNKEY)
          {
            player_0.y+=(player_0.movespeed+level/5);
          }
        }
        if(player_0.x < SPACE_X_MIN)
            player_0.x=SPACE_X_MAX;
        if(player_0.x>SPACE_X_MAX)
            player_0.x=SPACE_X_MIN;
    
        if(player_0.y < SPACE_Y_MIN)
            player_0.y=SPACE_Y_MAX;
        if(player_0.y>SPACE_Y_MAX)
            player_0.y=SPACE_Y_MIN;
    
        if(player_0.flash > 0)
            player_0.flash--;
        if(!(player_0.flash % 2))
            W_DrawImage
              (
                baseWin,
                player_0.x-(playerShip->width/2),
                player_0.y-(playerShip->height/2),
                counter, playerShip, W_Red
              );
        if(player_0.shield > 0)
            player_0.shield--;
        if(player_0.shield && ((player_0.shield > SHIELDTIME/4) || player_0.shield % 2)) {
            W_DrawImage
              (
                baseWin,
                (int)player_0.x-(shieldImage->width/2),
                player_0.y-(playerShip->height/2) - 3,
                0, shieldImage,W_Cyan
              );
        }
        else if(player_0.safety_bubbles)
        {
            W_DrawImage
              (
                baseWin,
                (int)player_0.x-(bubbleImage->width/2),
                player_0.y-(playerShip->height/2) - 3,
                0, bubbleImage,W_Cyan
              );
        }
    } else { /* paused */
        W_DrawImage
          (
            baseWin,
            (int)player_0.x-(playerShip->width/2),
            player_0.y-(playerShip->height/2),
            counter, playerShip, W_Red
          );
    }
}

int
main(argc, argv)
int     argc;
char  **argv;
{
    int ac;
    char *dpyname = 0;
    struct timeval tv, otv;
    long ut;
   
    gettimeofday(&otv, 0);
    for(ac = 1; ac < argc; ac++) {
        if(*argv[ac] == '-') {
            if(strcmp(argv[ac], "-scores") == 0) {
                print_scores();
                exit(0);
            }
            if(strcmp(argv[ac], "-display") == 0 && (ac+1 < argc)) {
                dpyname = argv[ac+1];
                ac++;
#if 0
            } else if (strcmp(argv[ac], "-b") == 0) {
                useBuffered = !useBuffered;
#endif
            } else if (strcmp(argv[ac], "-keyboard") == 0) {
                player_0.mouse_control = 0;
            } else if (strcmp(argv[ac], "-mouse") == 0) {
                player_0.mouse_control = 1;
#ifdef SOUND
            } else if (strcmp(argv[ac], "-nosound") == 0) {
                playSounds = 0;
#endif
            } else {
                print_usage();
                exit(0);
            }
        } else {
            print_usage();
            exit(0);
        }
    }
        
    W_Initialize(dpyname);

    backColor = W_Black;
    baseWin = W_MakeWindow("XGalaga", 0, 0, WINWIDTH, WINHEIGHT, 0, "tiny", 0, W_White);
    W_MapWindow(baseWin);

    /*    W_AutoRepeatOff();*/

    W_Flush();

    W_SetImageDir(IMAGEDIR);

    playerShip = getImage(I_PLAYER1);
    playerTorp = getImage(I_MTORP);
    enemyTorp = getImage(I_ETORP);
    shieldImage = getImage(I_SHIELD);
    bubbleImage= getImage(I_SAFETY_BUBBLE);

    level=1;

    load_scores();
    init_titles();
    init_player(&player_0,level);
    init_stars();
    init_explosions();
    init_score();
    init_prizes();
#ifdef SOUND
    init_sound();
#endif
    init_aliens(level);

    while(1) {
        counter++;
        if(gameOver && getting_name)
            undo_name();

        undo_score();
        undo_explosions();
        undo_stars();
        undo_aliens();
        undo_prizes();
        undo_etorps();
        undo_torps();
        /*        undo_clusters();*/
        undo_player();

        if(paused)
            undo_pause();

        W_FlushClearAreaCache(baseWin);

        if(paused)
            do_pause();

        do_stars();
        do_player();
        do_torps();
        /*        do_clusters();*/
        do_etorps();
        do_prizes();
        do_aliens();
        do_explosions();
        do_score();
        if(gameOver) {
            do_title();
            if(getting_name)
                do_name();
        }
        W_DisplayBuffer(baseWin);
        W_Flush();
        gettimeofday(&tv, 0);
        if(tv.tv_sec > otv.tv_sec)
            tv.tv_usec += 1000000;
        ut = tv.tv_usec - otv.tv_usec;
        if(UTIMER - ut > 0) {
            struct timeval tv;
            unsigned long usecs = UTIMER - ut;
            tv.tv_sec  = usecs / 1000000L;
            tv.tv_usec = usecs % 1000000L;
            (void) select (0, 0, 0, 0, &tv);
        }
        gettimeofday(&otv, 0);
    }
}
