#include <stdlib.h>
#include <stdio.h>
#include "libsprite/Wlib.h"
#include "images.h"
#include "struct.h"
#include "data.h"
#include "defs.h"
#include "proto.h"
#include "sound.h"

#define PRIZESPEED 3+player_0.prize_velocity

enum prize_type
{
  PR_SING,  
  PR_DOUB,  
  PR_TRIP,  
  PR_SPEED, 
  PR_SHIELD,
  PR_SMART, 
  PR_LEMON, 
  PR_EXTRABULLET,
  PR_WRAP,
  PR_SAFETY_BUBBLE,
  PR_CANNON_POWER,
  PR_SHOOT_DOWN_ETORPS,
  PR_INCREASE_TORP_VELOCITY,
  PR_DECREASE_TORP_VELOCITY,
  PR_INCREASE_PRIZE_VELOCITY,
  PR_DECREASE_PRIZE_VELOCITY,
  NUMPRIZES,
};
typedef enum prize_type prize_type_t;

W_Image *prizeImages[NUMPRIZES];
    
struct prize
{
  struct prize *next, *prev;
  int x, y, dying;
  int is_changing;
  prize_type_t type;
};
typedef struct prize prize_t;

extern struct torp *first_etorp;
extern int numetorps;

static int prize_image_map[NUMPRIZES]=
{
  I_PR_SING,
  I_PR_DOUB,
  I_PR_TRIP,
  I_PR_SPEED,
  I_PR_SHIELD,
  I_PR_BRAIN,
  I_PR_LEMON,
  I_PR_EXTRABULLET,
  I_PR_WRAP,
  I_PR_SAFETY_BUBBLE,
  I_PR_CANNON_POWER,
  I_PR_SHOOT_DOWN_ETORPS,
  I_PR_INCREASE_TORP_VELOCITY,
  I_PR_DECREASE_TORP_VELOCITY,
  I_PR_INCREASE_PRIZE_VELOCITY,
  I_PR_DECREASE_PRIZE_VELOCITY,
};

static prize_t *first_prize; /* 8-) */

void init_prizes()
{
  prize_type_t pr;

  for (pr= 0 ; pr < NUMPRIZES ; ++pr)
  {
    prizeImages[pr] = getImage(prize_image_map[pr]);
  }
}
    
void new_prize(double x, double y)
{
    prize_t *p;

    p = malloc(sizeof(prize_t));
    p->next = first_prize;
    p->prev = 0;
    if(first_prize)
	first_prize->prev = p;
    first_prize = p;
    p->x = x;
    p->y = y;
    p->type = random()%NUMPRIZES;
    p->dying = 0;
    p->is_changing= 0;
}

void undo_prizes()
{
  prize_t *p = first_prize, *nextp;

  while(p)
  {
    nextp = p->next;

    W_CacheClearArea
      (
        baseWin, 
        ((int)p->x)-prizeImages[p->type]->width/2, ((int)p->y)-prizeImages[p->type]->width/2,
        prizeImages[p->type]->width+1,
        prizeImages[p->type]->height
      );
    if(p->dying)
    {
      if(p->next)
      {
        p->next->prev = p->prev;
      }
      if(p->prev)
      {
        p->prev->next = p->next;
      }
      if(p == first_prize)
      {
        first_prize = p->next;
      }
      free(p);
    }
    p = nextp;
  }
}

void do_prizes()
{
    prize_t *p = first_prize;
    int i;
#ifdef SOUND
    int oldPlaySounds;
#endif

    while(p) {
	if(!paused)
	    p->y+=PRIZESPEED;
	W_DrawImage(baseWin, 
		    ((int)p->x)-prizeImages[p->type]->width/2, ((int)p->y)-prizeImages[p->type]->width/2,
		    0, prizeImages[p->type], W_Green);

	if
          (
            (!player_0.dead)&&
            (
              (p->x)>(player_0.x-(playerShip->width))&&
              (p->x)<(player_0.x+(playerShip->width))
            )&&
            (
              (p->y)>(player_0.y-(playerShip->height))&&
              (p->y)<(player_0.y+(playerShip->height))
            )&&
            (!p->dying)
          )
        {
	    p->dying = 1;
#ifdef SOUND
	    play_sound(SND_DDLOO);
#endif
            player_0.got_lemon=0;
	    switch(p->type) {
	      case PR_SING:
                player_0.maxtorps++;
                if(player_0.maxtorps>MAXTORPS)
                {
                  player_0.maxtorps=MAXTORPS;
                }
		break;
	      case PR_DOUB:
                player_0.maxtorps+=2;
		if(player_0.weapon<DOUBLESHOT){++player_0.weapon;}
		if(player_0.maxtorps < 4)
                {
                  player_0.maxtorps = 4;
                }
                if(player_0.maxtorps>MAXTORPS)
                {
                  player_0.maxtorps=MAXTORPS;
                }
		break;
	      case PR_TRIP:
                player_0.maxtorps+=3;
                if(player_0.weapon<TRIPLESHOT){player_0.weapon=TRIPLESHOT;}
		else {++player_0.weapon;}
                if(player_0.maxtorps<6)
                {
                  player_0.maxtorps=6;
                }
                if(player_0.maxtorps>MAXTORPS)
                {
                  player_0.maxtorps=MAXTORPS;
                }
		break;
	      case PR_SPEED:
		if(player_0.movespeed < MAXSPEED)
		    player_0.movespeed++;
		break;
	      case PR_SHIELD:
		player_0.shield = SHIELDTIME;
#ifdef SOUND
		play_sound(SND_SHIELD);
#endif
		break;
	      case PR_SMART:
#if !defined(ALLOW_SMART_BOMB)                
#ifdef SOUND
		play_sound(SND_SMART);
		oldPlaySounds = playSounds;
		playSounds = 0;
#endif
		for(i=0;i<MAXALIENS;i++)
                {
                  switch(aliens[i].state)
                  {
                    case WAITING_TO_ENTER:
                      /* Ha Ha. Aliens that haven't entered yet don't
                         get hurt.
                      */
                      break;
                    case ENTERING:
                    case IN_CONVOY:
                    case ATTACKING:
                    case ESCORTING:
                    case RETURNING_TO_CONVOY:
                      torp_hit_alien(aliens+i);
                    case DEAD:
                      /* Dead ones don't get hurt either.*/
                      break;
                    default:
                      fprintf(OUT_FILE,"Alien %i in unknown state %i!\n",i,aliens[i].state);
                  }
		}
#ifdef SOUND
		playSounds = oldPlaySounds;
#endif
#endif /* !defined(ALLOW_SMART_BOMB)                */
		break;
                
	      case PR_LEMON:
		player_0.got_lemon = 1;
		player_0.maxtorps = MINTORPS;
		player_0.weapon = 0;
		player_0.movespeed = MINSPEED;
                player_0.times_torps_can_wrap= 0;
                player_0.safety_bubbles= 0;
                player_0.cannon_power= 0;
                player_0.shoot_down_etorps= 0;
		break;
                
	      case PR_EXTRABULLET:
                /* Destroy all etorps.*/
                {
                  torp_t* current_torp= first_etorp;
                  if (current_torp)
                  {
                    do
                    {
                      current_torp->alive= 0;
                      current_torp= current_torp->next;
                      /*Remember, etorps is a circular linked list.*/
                    }while (current_torp && current_torp != first_etorp);
                  }
                }
		break;

              case PR_WRAP:
                if (player_0.times_torps_can_wrap < MAX_TIMES_TORPS_CAN_WRAP)
                {
                  ++player_0.times_torps_can_wrap;
                }
                if (player_0.maxtorps < MAXTORPS)
                {
                  player_0.maxtorps+=2;
                }
                break;

              case PR_SAFETY_BUBBLE:
                ++player_0.safety_bubbles;
                break;

              case PR_CANNON_POWER:
                ++player_0.cannon_power;
                player_0.maxtorps+=4;
                if(player_0.weapon<(NUMWEAPONS-2)){player_0.weapon+=2;}
                break;

              case PR_SHOOT_DOWN_ETORPS:
                ++player_0.shoot_down_etorps;
                ++player_0.torp_velocity;
                break;

              case PR_INCREASE_TORP_VELOCITY:
                ++player_0.torp_velocity;
                break;
                
              case PR_DECREASE_TORP_VELOCITY:
                --player_0.torp_velocity;
                break;

              case PR_INCREASE_PRIZE_VELOCITY:
                if(player_0.prize_velocity < 2)
                {
                  ++player_0.prize_velocity;
                }
                break;

              case PR_DECREASE_PRIZE_VELOCITY:
                if(player_0.prize_velocity > -8)
                {
                  --player_0.prize_velocity;
                }
                if(player_0.prize_velocity == -3)
                {
                  player_0.prize_velocity = -4;
                }
                break;
                
              default:
                fprintf(OUT_FILE,"Prize of  unknown type %i!\n",p->type);
	    }
	    if(player_0.maxtorps > MAXTORPS)
		player_0.maxtorps = MAXTORPS;
	} else if(p->y > WINHEIGHT) {
	    p->dying = 1;
	}
        else
        {
          if (p->is_changing)
          {
            if (!random()%2)
            {
              p->type= random()%NUMPRIZES;
            }
          }
        }
	p=p->next;
    }
}

